////////////////////////////////////////////////////////////////////// 
// pedscript/pedscript.cpp 
// (c) 2000-2010 Goncalo Abecasis
// 
// This file is distributed as part of the Goncalo source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile Goncalo.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday June 04, 2010
// 
 
#include "Pedigree.h"
#include "Random.h"
#include "Error.h"

#include <stdio.h>

Pedigree ped;
FILE * logfile = NULL;

void Log(const char * format, ...)
   {
   va_list args;
   va_start(args, format);

   vprintf(format, args);

   if (logfile) vfprintf(logfile, format, args);

   va_end(args);
   }

bool CheckArguments(StringArray & command, int arguments,
                   const char * syntax)
     {
     if (command.Length() == arguments)
        return true;

     printf("The command %s requires %d arguments.\n\n"
            "Syntax: %s\n\n", (const char *) command[0], arguments, syntax);

     return false;
     }

void ScriptLog(StringArray & command)
   {
   if (CheckArguments(command, 2, "LOGFILE filename"))
      {
      if (logfile != NULL)
         fclose(logfile);

      logfile = fopen(command[1], "wt");
      }
   }

void ScriptRead(StringArray & command)
   {
   const char * readSyntax = "READ [PED|DATA] filename";

   if (CheckArguments(command, 3, readSyntax))
      {
      if (command[1].MatchesBeginningOf("PEDIGREE") == 0)
         {
         ped.Load(command[2]);
         Log("## Loaded pedigree with %d individuals\n", ped.count);
         }
      else if (command[1].MatchesBeginningOf("DATA") == 0)
         {
         ped.Prepare(command[2]);
         Log("## Loaded data file with %d entries\n", ped.pd.columnCount);
         }
      else
         printf("%s\n\n", readSyntax);
      }
   }

void ScriptWrite(StringArray & command)
   {
   const char * writeSyntax = "WRITE [PEDIGREE|DATA] filename";

   if (CheckArguments(command, 3, writeSyntax))
      {
      if (command[1].MatchesBeginningOf("DATA") == 0)
         {
         ped.WriteDataFile(command[2]);
         Log("## Wrote data file\n");
         }
      else if (command[1].MatchesBeginningOf("PEDIGREE") == 0)
         {
         ped.WritePedigreeFile(command[2]);
         Log("## Wrote pedigree file\n");
         }
      else
         printf("%s\n\n", writeSyntax);
      }
   }

void ScriptMendelCheck()
   {
   bool success = ped.InheritanceCheck(false);

   if (success) Log("## No errors detected\n");
   }

void ScriptHelp()
   {
   printf("The available commands are:\n\n"
          "  ANONYMIZE     -- Remove individual identifiers from pedigree\n"
          "  READ | WRITE  -- To input and output pedigree and data files\n"
          "  TRIM          -- Remove uninformative individuals\n"
          "  NUKE          -- Break all family links, marking everyone as unrelated\n"
          "  SAMPLE        -- Sample individuals or families to generate new pedigree\n"
          "  SET           -- Change value of specific traits or covariates\n"
          "  MENDELCHECK   -- Run Mendelian Inheritance Check\n"
          "  QUIT          -- End pedscript\n"
          "  HELP          -- Print this help\n\n"
          "You can get details of any command by executing it with no arguments\n");
   printf("You are using the PedScript interpreted language, version 0.03\n");
   }

#define LESS_THAN    1
#define EQUAL_TO     2
#define GREATER_THAN 3
#define NOT_EQUAL_TO 4

int GetCondition(String & condition)
   {
   if (condition == ">")
      return GREATER_THAN;
   else if (condition == "<")
      return LESS_THAN;
   else if (condition == "==" || condition == "=" || condition.MatchesBeginningOf("IS") == 0)
      return EQUAL_TO;
   else if (condition == "!=")
      return NOT_EQUAL_TO;
   else
      return -1;
   }

bool CheckCondition(int condition, double value1, double value2)
   {
   switch (condition)
      {
      case GREATER_THAN :
         return value1 > value2 && value1 != _NAN_ && value2 != _NAN_;
      case LESS_THAN :
         return value1 < value2 && value1 != _NAN_ && value2 != _NAN_;
      case EQUAL_TO :
         return value1 == value2;
      case NOT_EQUAL_TO :
         return value1 != value2;
      }
   return false;
   }

#define REF_MASK        0xF0000000
#define REF_TRAIT       0x10000000
#define REF_COVARIATE   0x20000000
#define REF_FEATURE     0x30000000
#define REF_MARKER      0x40000000

int GetReference(String & what, String & label)
   {
   if (what.MatchesBeginningOf("TRAIT") == 0)
      return REF_TRAIT | ped.LookupTrait(label);
   else if (what.MatchesBeginningOf("COVARIATE") == 0)
      return REF_COVARIATE | ped.LookupCovariate(label);
   else if (what.MatchesBeginningOf("MARKER") == 0)
      return REF_MARKER | ped.LookupMarker(label);
   return -1;
   }

double GetValue(String & value)
   {
   if (value.MatchesBeginningOf("NAN") == 0 || value.MatchesBeginningOf("MISSING") == 0)
      return _NAN_;
   else if (value.MatchesBeginningOf("HOMOZYGOUS") == 0)
      return -1.0;
   else if (value.MatchesBeginningOf("HETEROZYGOUS") == 0)
      return -2.0;
   else
      return atof(value);
   }

double & Retrieve(Person & p, int reference)
   {
   if ((reference & REF_MASK) == REF_TRAIT)
      return p.traits[reference & ~REF_MASK];
   else if ((reference & REF_MASK) == REF_COVARIATE)
      return p.covariates[reference & ~REF_MASK];
   else if ((reference & REF_MASK) == REF_MARKER)
      {
      static double dummy_genotype = _NAN_;

      dummy_genotype =
         !p.markers[reference & ~REF_MASK].isKnown() ? _NAN_ :
          p.markers[reference & ~REF_MASK].isHeterozygous() ? -2.0 : -1.0;

      return dummy_genotype;
      }

   // If we get here there has been an error
   error("Attempt to retrieve unknown feature");
   return *(new double);
   }

void ScriptSet(StringArray & command)
   {
   const char * setSyntax = "SET [TRAIT|COVAR] label TO value WHEN [TRAIT|COVAR|MARKER] label [COMPARISON] value\n";

   if (CheckArguments(command, 10, setSyntax))
      {
      int condition = GetCondition(command[8]);
      int output = GetReference(command[1], command[2]);
      int input = GetReference(command[6], command[7]);
      double value = GetValue(command[4]);
      double test = GetValue(command[9]);

      if (command[3].MatchesBeginningOf("TO") != 0 ||
          command[5].MatchesBeginningOf("WHEN") != 0 ||
          output == -1 || input == -1 || condition == -1)
         {
         printf("%s\n\n", setSyntax);
         return;
         }

      int changed = 0;

      for (int i = 0; i < ped.count; i++)
         if (CheckCondition(condition, Retrieve(ped[i], input), test))
            {
            Retrieve(ped[i], output) = value;
            changed++;
            }

      Log("## Changed %d values\n", changed);
      }
   }

void ScriptTrim(StringArray & )
   {
   int individuals = ped.count;
   int families = ped.familyCount;

   ped.Trim(true);

   Log("# Dropped %d of %d individuals\n", individuals - ped.count, individuals);
   Log("# Dropped %d of %d families\n", families - ped.familyCount, families);
   }

void ScriptAnonymize(StringArray & )
   {
   for (int i = 0; i < ped.familyCount; i++)
      for (int j = ped.families[i]->first; j <= ped.families[i]->last; j++)
         {
         ped[j].famid = i + 1;
         ped[j].pid = ped[j].traverse + 1;

         ped[j].motid = ped[j].mother == NULL ? 0 : ped[j].mother->traverse + 1;
         ped[j].fatid = ped[j].father == NULL ? 0 : ped[j].father->traverse + 1;
         }

   ped.Sort();

   Log("# Anonymized pedigree with %d inviduals\n", ped.count);
   }

void ScriptNuke(StringArray &)
   {
   int families = ped.familyCount;

   for (int i = 0; i < ped.count; i++)
      ped[i].famid = ped[i].famid + "->" + ped[i].pid;

   ped.Sort();

   Log("# %d inviduals extracted from %d families\n", ped.count, families);
   }

void ScriptSample(StringArray & words)
   {
   const char * sampleSyntax = "SAMPLE number [FOUNDERS|FAMILIES|PERSONS] TO file";

   if (CheckArguments(words, 5, sampleSyntax))
      {
      int number = words[1];
      FILE * output = fopen(words[4], "wt");

      if (words[3].MatchesBeginningOf("TO") != 0 || words[2] == "F" || words[2] == "f")
         {
         printf("%s\n\n", sampleSyntax);
         return;
         }

      if (number < 0) number = 0;

      if (words[2].MatchesBeginningOf("FOUNDERS") == 0)
         {
         double population = 0;

         for (int i = 0; i < ped.count; i++)
            if (ped[i].isFounder())
               population++;

         int count = number > population ? population : number;

         for (int i = 0; i < ped.count; i++)
            if (ped[i].isFounder())
               {
               if (number / population >= globalRandom.Next())
                  {
                  ped.WritePerson(output, i);
                  number--;
                  }

               population--;
               }

         printf("# SAMPLED %d of %d FOUNDERS TO %s\n", count, population, (const char *) words[4]);
         }

      if (words[2].MatchesBeginningOf("PERSONS") == 0 || words[2].MatchesBeginningOf("INDIVIDUALS") == 0)
         {
         int count = number > ped.count ? ped.count : number;

         for (int i = ped.count; i >= 0; i--)
            if (number / (double) (i + 1) >= globalRandom.Next())
               {
               ped.WritePerson(output, i);
               number--;
               }

         printf("# SAMPLED %d of %d INDIVIDUALS TO %s\n", count, ped.count, (const char *) words[4]);
         }

      if (words[2].MatchesBeginningOf("FAMILIES") == 0)
         {
         int count = number > ped.familyCount ? ped.familyCount : number;

         for (int i = ped.familyCount - 1; i > 0; i--)
            if (number / (double) (i + 1) >= globalRandom.Next())
               {
               for (int j = ped.families[i]->first; j <= ped.families[j]->last; j++)
                  ped.WritePerson(output, j);

               number--;
               }

         printf("# SAMPLED %d of %d FAMILIES TO %s\n", count, ped.familyCount, (const char *) words[4]);
         }

      fclose(output);
      }
   }

int main(int argc, char ** argv)
   {
   String::caseSensitive = false;

   printf("PedScript - (c) 2004 Goncalo Abecasis\n"
          "            Scripting language for manipulation of pedigree files\n\n");

   String      line;
   StringArray words;

   while (!feof(stdin))
      {
      line.ReadLine();
      words.Clear();
      words.AddTokens(line);

      if (logfile != NULL)
         fprintf(logfile, "%s\n", (const char *) line);

      if (words.Length() == 0 || words[0][0] == '#' || words[0] == "//")
         continue;

      if (words[0].MatchesBeginningOf("ANONYMIZE") == 0)
         ScriptAnonymize(words);
      else if (words[0].MatchesBeginningOf("HELP") == 0)
         ScriptHelp();
      else if (words[0].MatchesBeginningOf("LOG") == 0)
         ScriptLog(words);
      else if (words[0].MatchesBeginningOf("MENDELCHECK") == 0)
         ScriptMendelCheck();
      else if (words[0].MatchesBeginningOf("NUKE") == 0)
         ScriptNuke(words);
      else if (words[0].MatchesBeginningOf("QUIT") == 0)
         break;
      else if (words[0].MatchesBeginningOf("READ") == 0)
         ScriptRead(words);
      else if (words[0].MatchesBeginningOf("SET") == 0 && words[0].Length() >= 2)
         ScriptSet(words);
      else if (words[0].MatchesBeginningOf("SAMPLE") == 0 && words[0].Length() >= 2)
         ScriptSample(words);
      else if (words[0].MatchesBeginningOf("TRIM") == 0)
         ScriptTrim(words);
      else if (words[0].MatchesBeginningOf("WRITE") == 0)
         ScriptWrite(words);
      else
         Log("# Unrecognized command\n");
      }

   if (logfile != NULL)
      fclose(logfile);
   }

 
